"""
"""
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip
import math

__all__ = (
    'equivalent',
    'equivalent_complex',
    'equivalent_matt',
    'equivalent_matrix',
    'equivalent_sequence',
    'show_complex_difference',
    'TOL',
)

TOL = 1E-13
#-----------------------------------------------------
def equivalent_sequence(u,m,tol=TOL):
    """
    Compare each element of a pair of sequences
    for equivalence
    
    """
    assert len(u) == len(m), "different lengths: %s and %s" % (u,m)
    OK = True
    for u_i,m_i in izip(u,m):
        OK &=  abs(u_i - m_i) < tol       

    if OK:
        return True
    else:
        print("Tolerance = %G" % tol)
        print("Differences: ")
        for u_i,m_i in izip(u,m):
            print(abs(u_i - m_i))
            
        raise AssertionError("'%s' <> '%s' " % (u,m))

#-----------------------------------------------------
def equivalent_matt(m,n,tol=TOL):
    """
    `m` and `n` can be mapping objects or namedtuples
    
    """
    # convert namedtuples to mappings
    if hasattr(m,'_asdict'):
        m = m._hasdict()
    if hasattr(n,'_asdict'):
        n = n._hasdict()
        
    keys = m.keys()
    if (keys == n.keys()):
        for k in keys:
            if abs(m[k] - n[k]) >= tol:
                print("Values for '%s' differ: abs(%.15G - %.15G) = %.15G" % (
                        k,
                        m[k],
                        n[k],
                        abs(m[k] - n[k])
                    ))
                return False
        return True
    else:
        return False
    
#-----------------------------------------------------
def equivalent_matrix(u,m,tol=TOL):
    if(
        abs(u[0,0] - m[0,0]) < tol
    and abs(u[1,0] - m[1,0]) < tol
    and abs(u[0,1] - m[0,1]) < tol
    and abs(u[1,1] - m[1,1]) < tol
    ):
        return True
    else:
        print("Differences and tolerance: {} {} {} {} {}".format(
            abs(u[0,0] - m[0,0])
        ,   abs(u[1,0] - m[1,0])
        ,   abs(u[0,1] - m[0,1])
        ,   abs(u[1,1] - m[1,1])
        ,   tol))
        raise AssertionError("'%s' <> '%s' " % (u,m))
#-----------------------------------------------------
def show_complex_difference(x,y):
    """-> a string showing the difference in each component
    """
    z = complex(x)-complex(y)
    return "Differences: (%.15G,%.15G)" %(z.real, z.imag)
    
#-----------------------------------------------------
def equivalent_complex(x,y,tol=TOL):
    "Test the numerical equivalence of the complex arguments"
    xc = complex(x)
    yc = complex(y)
    if _equivalent(xc.real,yc.real,tol) and _equivalent(xc.imag,yc.imag,tol):
        return True
    else:
        print("Differences and tolerance: {} {} {}".format(abs(xc.real-yc.real), abs(xc.imag-yc.imag), tol))
        raise AssertionError("'(%.15G,%.15G)' <> '(%.15G,%.15G)' " % (x.real,x.imag,y.real,y.imag))

#-----------------------------------------------------
def _equivalent(x,y,tol):
    if( math.isinf(x) and math.isinf(y) and x == y ):
        return True
    elif( abs(x-y) < tol  ):
        return True
    else:
        return False
    
def equivalent(x,y,tol=TOL):
    """Test the numerical equivalence of the arguments
    """
    if _equivalent(x,y,tol):
        return True
    else:
        msg = "Values are not numerically equivalent: abs(%.16G-%.16G) = %.16G with tol= %.16G" % (
            x,y,abs(x-y),tol)
        raise AssertionError(msg)



import unittest

from GTC import *

from testing_tools import *

TOL = 1E-13 
 
#-----------------------------------------------------
class TestMisc(unittest.TestCase):

    def test_illegal_u_component_cases(self):
        # Cannot ask for ...
        
        # non-intermediate real component
        x1 = ureal(1,1) 
        x2 = x1 + ureal(2,1)
        y = x1**x2
        self.assertRaises(RuntimeError,rp.u_component,y,x2)
        
        # non-intermediate complex component
        x1 = ucomplex(10,1)
        x2 = x1**2 
        y = magnitude(x1*3j + x2) 
        self.assertRaises(TypeError,rp.u_component,y,x2)
        
        # non-intermediate complex component
        y = x1*3j + x2
        self.assertRaises(RuntimeError,rp.u_component,y,x2)
 
        # non-intermediate real component 
        x1 = ureal(10,1)
        x2 = x1**2 
        y = x1*3j + x2 
        self.assertRaises(TypeError,rp.u_component,y,x2)
        
        # y must be an uncertain type
        self.assertRaises(RuntimeError,rp.u_component,10.0,x2)
 
    def test_simple_cases(self):
        # Return zero when ...
        
        # real y, uncertain complex constant 
        x1 = ureal(10,1)
        x2 = constant(1+5j) 
        self.assertTrue( equivalent_sequence(
            rp.u_component(x1,x2),
            (0.,0.,0.,0.),
            TOL
        ))
        
        # real y, complex-valued x 
        y = ureal(1,1) 
        self.assertTrue( equivalent_sequence(
            rp.u_component(y,1+7j),
            (0.,0.,0.,0.),
            TOL
        ))
        
        # complex y, uncertain complex constant 
        x1 = ucomplex(10,1)
        x2 = constant(1+5j) 
        self.assertTrue( equivalent_sequence(
            rp.u_component(x1,x2),
            (0.,0.,0.,0.),
            TOL
        ))
        
        
#-----------------------------------------------------
class TestInCoverage(unittest.TestCase):
    """
    Functions that take a coverage factor and refurn dof
    """
    
    def test_k_to_dof(self):
        """
        1-D cases
        """
        for p in (90,95,99):
            for df in (1,5,10,25,55,1001):
                k = rp.k_factor(df,p)
                dof = rp.k_to_dof(k,p)
                self.assertTrue( equivalent(df,dof,1E10) )
            

        # Illegal cases
        k = -16.58005817 
        self.assertRaises(RuntimeError,rp.k_to_dof,k)
        
        k = rp.k_factor(1.9)
        self.assertRaises(RuntimeError,rp.k_to_dof,k,p=100)
        self.assertRaises(RuntimeError,rp.k_to_dof,-k)
   
    def test_k2_to_dof(self):
        """
        2-D cases
        """
        for p in (90,95,99):
            for df in (2,5,10,25,55,1001):
                k = math.sqrt( rp.k2_factor_sq(df,p) )
                dof = rp.k2_to_dof(k,p)
                self.assertTrue( equivalent(df,dof,1E10) )
 
        # Illegal cases
        k = math.sqrt( rp.k2_factor_sq(2) ) + 0.1
        self.assertRaises(RuntimeError,rp.k2_to_dof,k)
        
        k = math.sqrt( rp.k2_factor_sq(2) )
        self.assertRaises(RuntimeError,rp.k2_to_dof,k,p=100)
        self.assertRaises(RuntimeError,rp.k2_to_dof,-k)
  
#-----------------------------------------------------
class TestCoverage(unittest.TestCase):
    
    """
    The tests are made for 95% and 99% coverage using
    values calculated to 10 digits by R-2.9.2
    
    """
           
    def test_k_factor(self):
        TOL = 1E-7
        data_95 = (
            (2,4.302652730),(3,3.182446305),(5,2.570581836), (10,2.228138852), (50,2.008559112), (inf,1.959963985)
        )
        self.assertTrue(
            equivalent(data_95[0][1],rp.k_factor(data_95[0][0],95),1E-2 ) 
        )
        self.assertTrue(
            equivalent(data_95[1][1],rp.k_factor(data_95[1][0],95),1E-3 ) 
        )
            
        for df,k in data_95:
            self.assertTrue( equivalent(k,rp.k_factor(df),TOL) )

        data_99 = (
            (2,9.924843201),(3,5.84090931),(5,4.032142984), (10,3.169272673), (50,2.677793271), (inf,2.575829304)
        )
        self.assertTrue(
            equivalent(data_99[0][1],rp.k_factor(data_99[0][0],99),0.2 ) 
        )
        self.assertTrue(
            equivalent(data_99[1][1],rp.k_factor(data_99[1][0],99),0.02 ) 
        )
        for df,k in data_99:
            self.assertTrue( equivalent(k,rp.k_factor(df,99),TOL ) )
            
        # illegal `df`
        self.assertRaises(RuntimeError,rp.k_factor,0.5)
        
    def test_k2_factor(self):
        TOL = 1E-7
        data_95 = (
           (2,28.248893784), (5,4.166614906), 
            (10,3.075528764), (50,2.550142994), 
            (inf,2.447746831),
        )
        # illegal `df`
        self.assertRaises(RuntimeError,rp.k2_factor_sq,1)
        
        self.assertTrue(
            equivalent(data_95[1][1]**2,rp.k2_factor_sq(data_95[1][0],95),0.2 ) 
        )
        self.assertTrue(
            equivalent(data_95[2][1]**2,rp.k2_factor_sq(data_95[2][0],95),0.1 ) 
        )
        self.assertTrue(
            equivalent(data_95[3][1]**2,rp.k2_factor_sq(data_95[3][0],95),0.1 ) 
        )
        self.assertTrue(
            equivalent(data_95[4][1]**2,rp.k2_factor_sq(data_95[4][0],95),0.01 ) 
        )

        for df,k in data_95:
            self.assertTrue(
                equivalent(
                    k,
                    math.sqrt( 
                        reporting.k2_factor_sq(df)
                    ),
                    TOL) 
        )

        data_99 = (
           (2,141.414284993), (5,6.708203932), (10,4.222036715), (50,3.215529821), (inf,3.034854259)               
        )
        
        self.assertTrue(
            equivalent(data_99[1][1]**2,rp.k2_factor_sq(data_99[1][0],99),1.2 ) 
        )
        self.assertTrue(
            equivalent(data_99[2][1]**2,rp.k2_factor_sq(data_99[2][0],99),0.6 ) 
        )
        self.assertTrue(
            equivalent(data_99[3][1]**2,rp.k2_factor_sq(data_99[3][0],99),0.4 ) 
        )

        for df,k in data_99:
            self.assertTrue(
                equivalent(
                    k,
                    math.sqrt( 
                        reporting.k2_factor_sq(df,99)
                    ),
                    TOL
                ) 
            )          
 
#-----------------------------------------------------
class TestBudget(unittest.TestCase):
    
    def test_errors(self):
        
        x = ureal(1,1,label="x")
        y = ureal(1,1,label="y")
        z = ureal(1,1,label="z")

        x1 = result(x + y,label='x1')
        y1 = result(y + z,label='y1')

        x2 = x1 + y1

        # Illegal combination of keywords
        self.assertRaises(
            RuntimeError,rp.budget,x2,influences=[x1],intermediate=True
        )
        
    def test_empty_budget(self):
        """
        When there are no influence quantities, an empty 
        sequence should be returned. 
        
        """
        seq = rp.budget(3.0)
        self.assertEqual(len(seq),0)
 
        x = constant(6)
        seq = rp.budget(x)
        self.assertEqual(len(seq),0)
 
        x = constant(6+4j)
        seq = rp.budget(x)
        self.assertEqual(len(seq),0)
        x = constant(6)
        seq = rp.budget(x)
        self.assertEqual(len(seq),0)
 
        x = constant(6+4j)
        seq = rp.budget(x)
        self.assertEqual(len(seq),0)

    def test_limited_budget(self):
        x1 = ureal(0,1)
        x2 = ureal(0,1)
        x3 = ureal(0,1)
        y = x1 + x2 + x3 
        
        seq = rp.budget(y)
        self.assertEqual( len(seq), 3 )
        seq = rp.budget(y,max_number=2)
        self.assertEqual( len(seq), 2 )
        
    def test_real(self):
        """
        The budget of a real quantity will consist of a
        list of named tuples with elements for the
        labels and values (the magnitude)
        of the components of uncertainty.

        When a complex quantity has been involved, expect to
        see the components of uncertainty in terms of the
        real and imaginary components.

        A sequence of influences may be given, which may include
        real or complex uncertain numbers.
        
        """
        x1 = dict(x=1,u=.1,label='x1')
        x2 = dict(x=2,u=.2,label='x2')

        # defined as a complex but it has 0 imaginary        
        z1 = dict(z=3,u=1,label='z1')    
        
        ux1 = ureal(**x1)
        ux2 = ureal(**x2)
        uz1 = ucomplex(**z1)

        y = -ux1 + ux2 * magnitude(-uz1)

        # Trim should remove the zero element b[3]
        b = reporting.budget(y,trim=0.01)
        self.assertEqual( len(b), 3  )

        b = reporting.budget(y,trim=0)
        self.assertEqual( len(b), 4  )
        
        # The default order is in terms of the biggest uncertainty
        self.assertEqual(b[0].label,'z1_re')
        self.assertEqual(b[1].label,'x2')
        self.assertEqual(b[2].label,'x1')
        self.assertEqual(b[3].label,'z1_im')

        self.assertTrue( equivalent(b[0].u,2.0,TOL) )
        self.assertTrue( equivalent(b[1].u,3*.2,TOL) )
        self.assertTrue( equivalent(b[2].u,0.1,TOL) )
        self.assertTrue( equivalent(b[3].u,0,TOL) )

        # Sorting in different ways
        b = reporting.budget(y,reverse=False,trim=0)
        self.assertEqual(b[0].label,'z1_im')
        self.assertEqual(b[1].label,'x1')
        self.assertEqual(b[2].label,'x2')
        self.assertEqual(b[3].label,'z1_re')
        
        b = reporting.budget(y,key='label',reverse=False,trim=0)
        self.assertEqual(b[0].label,'x1')
        self.assertEqual(b[1].label,'x2')
        self.assertEqual(b[2].label,'z1_im')
        self.assertEqual(b[3].label,'z1_re')

        # With triming
        b = reporting.budget(y,key='label',reverse=False,trim=0.01)
        self.assertEqual(b[0].label,'x1')
        self.assertEqual(b[1].label,'x2')
        self.assertEqual(b[2].label,'z1_re')

        b = reporting.budget(y,influences=[ux1,uz1.real,uz1.imag],trim=0)
        self.assertEqual( len(b), 3  )
        self.assertEqual(b[0].label,'z1_re')
        self.assertEqual(b[1].label,'x1')
        self.assertEqual(b[2].label,'z1_im')

        # A complex quantity may be passed as an
        # influence but the budget reports real
        # and imaginary components
        b = reporting.budget(y,influences=[ux1,uz1],trim=0)
        self.assertEqual( len(b), 3  )
        self.assertEqual(b[0].label,'z1_re')
        self.assertEqual(b[1].label,'x1')
        self.assertEqual(b[2].label,'z1_im')

    def test_complex(self):
        """
        The budget of a complex quantity will consist of a
        list of named tuples with elements for the
        labels and values (the u_bar magnitude)
        of the components of uncertainty.

        Real quantities can be involved.

        Sorting is tested in the real case above.        
        
        """
        z1 = dict(z=1+1j,u=(1,1),label='z1')
        uz1 = ucomplex(**z1)
        
        z2 = dict(z=2-1j,u=(.5,.5),label='z2')
        uz2 = ucomplex(**z2)
        
        x1 = dict(x=1,u=.1,label='x1')
        ux1 = ureal(**x1)

        y = -uz1 + uz2* ux1
        
        b = reporting.budget(y)
        self.assertEqual( len(b), 3)

        self.assertTrue( equivalent(b[0].u,1.0,TOL) )
        self.assertTrue( equivalent(b[1].u,0.5,TOL) )
        self.assertTrue( equivalent(b[2].u,math.sqrt((.1**2 + .2**2)/2),TOL) )

        b = reporting.budget(y,influences=[ux1,uz1])
        self.assertEqual( len(b), 2)
        
        self.assertTrue( equivalent(b[0].u,1.0,TOL) )
        self.assertTrue( equivalent(b[1].u,math.sqrt((.1**2 + .2**2)/2),TOL) )

    def test_intermediate_real(self):
        
        x = ureal(1,1,label="x")
        y = ureal(1,1,label="y")
        z = ureal(1,1,label="z")

        x1 = result(x + y,label='x1')
        y1 = result(y + z,label='y1')

        # First possibility: undeclared result 
        x2 = x1 + y1
        
        # NB, expect 2 items; 3 would be an error
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )
 
        # Second possibility: declared result, no label
        x2 = result(x1 + y1)
        
        # NB, expect 2 items. 
        # Here 3 is a possible fault 
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )
 
        # Third possibility: declared result
        result(x1 + y1,label='x2')
        
        # NB, expect 2 items. 
        # Here 3 is a possible fault 
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )

    def test_intermediate_complex(self):
        
        x = ucomplex(1,1,label="x")
        y = ucomplex(1,1,label="y")
        z = ucomplex(1,1,label="z")

        x1 = result(x + y,label='x1')
        y1 = result(y + z,label='y1')

        # First possibility: undeclared result 
        x2 = x1 + y1
        
        # NB, expect 2 items; 3 would be an error
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )
 
        # Second possibility: declared result, no label
        x2 = result(x1 + y1)
        
        # NB, expect 2 items. 
        # Here 3 is a possible fault 
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )
 
        # Third possibility: declared result
        result(x1 + y1,label='x2')
        
        # NB, expect 2 items. 
        # Here 3 is a possible fault 
        x1_,y1_ = rp.budget(x2,intermediate=True)
        
        self.assertEqual( x1_.label, x1.label)
        self.assertEqual( y1_.label, y1.label)
        
        self.assertTrue( equivalent(
            x1_.u,
            math.sqrt(2),
            TOL
        ) )
  
        self.assertTrue( equivalent(
            y1_.u,
            math.sqrt(2),
            TOL
        ) )    

   
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest
import math

TOL = 1E-13
DIGITS = 13

from GTC import *
from testing_tools import *

implicit = fn.implicit

#-----------------------------------------------------
class TestImplicit(unittest.TestCase):
    """
    The implicit(fn,x_min,x_max,epsilon=1E-10) function returns an UncertainReal
    number 'x' that satisfies abs( fn(x) ) < epsilon.

    A RuntimeError will be raised if the root search does not converge.
    
    An exception will be raised is:
        * fn(x) does not return an UncertainReal when `x` is UncertainReal
        * [x_min,x_max] does not appear to include a zero (this is done by
            checking that fn(x_max) * fn(x_min) < 0)
        * x_min < x_max
    
    """
    def test_preconditions(self):
        fn = lambda x : x + ureal(0,1)
        self.assertRaises(RuntimeError,implicit,fn,10,1)
        self.assertRaises(RuntimeError,implicit,fn,10,20)
        
        # This will convert an uncertain number `x` into a float.
        fn = lambda x : value(x)
        self.assertRaises(AssertionError,implicit,fn,-1,1)

    def test_root_find(self):
        # solve 0 = cos(x), this has no uncertainty!
        fn = lambda x: cos(x)   
        x = implicit(fn,0,math.pi)
        self.assertTrue( equivalent(value(x), math.pi/2.0 ) )
        self.assertTrue( equivalent( value(fn(x)),0 ) )
        self.assertTrue( equivalent( rp.u_component(x,x),0.0 ) )

        # solve 0 = (x-x1) - y
        y = ureal(0.5,.1)
        x1 = ureal(1,1)
        fn = lambda x: (x-x1) - y
        x = implicit(fn,0.5,2.0)
        self.assertTrue( equivalent( value(x), 1.5 ) )
        self.assertTrue( equivalent( value(fn(x)),0 ) )
        self.assertTrue( equivalent( rp.u_component(x,x1),1.0 ) )
        self.assertTrue( equivalent( rp.u_component(x,y),.1 ) )
        self.assertTrue( equivalent( uncertainty(x), math.sqrt(variance(y)+variance(x1))) )
  
#===============================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'
    
    


import unittest
import os
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip

from GTC import *
from GTC.context import Context
from GTC import context
from GTC import persistence
from GTC.vector import is_ordered

from testing_tools import *

TOL = 1E-13 

#-----------------------------------------------------
class TestArchive(unittest.TestCase):

    def test_context_management(self):
        """
        Context registers should be updated when all dependence
        on an influence quantity disappears.
        
        """
        context._context = Context()
        
        x_seq = [1,1,1]
        u_seq = [1,1,1]
        label_seq = ['x1', 'x2', 'x3']
        x1, x2, x3 = multiple_ureal(x_seq,u_seq,df=5,label_seq=label_seq)

        set_correlation(.5,x1,x2)
        set_correlation(.5,x2,x3)
        
        x4 = result(x1 + x2)
        x5 = result(x2 + x3)
        x6 = result(x4 + x5)

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack and test
        context._context = Context()
        ar._thaw()
        
        y1, y2, y3 = ar.extract('x1','x2','x3')
        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertTrue( equivalent(
            get_correlation(x1,x2),
            get_correlation(y1,y2),
            TOL
        ))

        self.assertEqual( repr(x1), repr(y1) )
        self.assertEqual( repr(x2), repr(y2) )
        self.assertEqual( repr(x3), repr(y3) )
        self.assertEqual( repr(x4), repr(y4) )
        self.assertEqual( repr(x5), repr(y5) )
        self.assertEqual( repr(x6), repr(y6) )

        # self.assertTrue ( y1._node in c2._correlations._mat )
        # self.assertTrue ( y2._node in c2._correlations._mat )
        # self.assertTrue ( y3._node in c2._correlations._mat )

        # del y6
        
        # self.assertTrue ( y1._node in c2._correlations._mat )
        # self.assertTrue ( y2._node in c2._correlations._mat )
        # self.assertTrue ( y3._node in c2._correlations._mat )

        # del y3,y5

        # self.assertTrue ( y1._node in c2._correlations._mat )
        # self.assertTrue ( y2._node in c2._correlations._mat )

        # del y2, y4      

        # self.assertTrue ( y1._node in c2._correlations._mat )

        # del y1

    def test_errors(self):
        """
        Error conditions:
##            - trying to archive a complex and one of its components, or vice versa
##            - trying to archive the same object with different names
            - trying to extract with the wrong name
            - adding to an archive that has been frozen
            - extracting from an archive that has not been thawed
            - restoring an archive to the same context
            
        """
        context._context = Context() 
        
        x = ureal(1,1)
        y = ureal(2,1)
        z = result( x + y )

        ar = persistence.Archive()

        ar.add(x=x,z=z)
        self.assertRaises(RuntimeError,ar.add,x=y)  # same name

        # These used to be illegal but are now allowed
        zc = ucomplex(1,1)
        ar.add(zc=zc)
        ar.add(zr=zc.real)  # a component
        
        zz = ucomplex(1,1)
        ar.add(zr2=zz.real)
        ar.add(zz=zz)  # a component was archived

        ar._freeze()

        self.assertRaises(RuntimeError,ar.extract,'x')  # cannot extract yet
        
        context._context = Context()
        ar._thaw( )

        self.assertRaises(RuntimeError,ar.add,y=y)  # cannot add now
  
    def test_empty_archive(self):
        "Nothing to save is surely an error?"
        ar = persistence.Archive() 
        self.assertRaises(RuntimeError,ar._freeze) 
 
    def test_forgot_result(self):
        """
        Raise a RuntimeError not an AttributeError when 
        someone forgets to declare an intermediate result 
        using `result()`
        
        """
        context._context = Context() 
        
        x = ureal(1,1)
        y = ureal(2,1)
        z = ( x + y )

        ar = persistence.Archive()
        self.assertRaises(RuntimeError,ar.add,z=z)
 
        x = ucomplex(1,1)
        y = ucomplex(2,1)
        z = ( x + y )
        self.assertRaises(RuntimeError,ar.add,z=z)

 
    def test(self):
        """
        Simple x,y z problem, but don't save
        one of the elementary uns.
        
        """
        context._context = Context()
        
        x = ureal(1,1)
        y = ureal(2,1)
        z = result( x + y )

        self.assertTrue( z.is_intermediate )

        ar = persistence.Archive()

        ar.add(x=x,z=z)
        ar._freeze()
                
        # Unpack and check 
        context._context = Context()
        ar._thaw()

        # Single argument unpacks to an object
        # not a sequence
        x1 = ar.extract('x')
        z1 = ar.extract('z')

        self.assertTrue( z1.is_intermediate )
        self.assertTrue(
            all( 
                z_i.uid == z1_i.uid 
                    for (z_i, z1_i) in izip(
                        z._i_components,z1._i_components) 
            )
        )
        a = component(z,x)
        b = component(z1,x1)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vectors are well-formed
        self.assertTrue( is_ordered(z1._u_components) )
        self.assertTrue( is_ordered(z1._d_components) )
        self.assertTrue( is_ordered(z1._i_components) )
        
    def test_attributes(self):
        """
        Dict-like attributes: keys(), values() items()
        and the 'iter' variants, also len()
        
        """
        context._context = Context() 
        
        x = ureal(1,1)
        y = ureal(2,1)
        z = result(x + y)

        ar = persistence.Archive()
        self.assertEqual( len(ar), 0)
        
        ar.add(x=x,z=z)
        self.assertEqual( len(ar) , 2)

        names = sorted(['x','z'])
        objs = sorted([x,z])
        self.assertEqual( names,sorted(ar.keys()) )
        self.assertEqual( objs,sorted(ar.values()) )
        self.assertEqual( sorted(zip(names,objs)),sorted(ar.items()) )

        for i,k in enumerate( sorted(ar.iterkeys()) ):
            self.assertEqual( names[i], k )

        for i,k in enumerate( sorted(ar.itervalues()) ):
            self.assertEqual( objs[i], k )

        items = sorted(zip(names,objs))
        for i,k in enumerate( sorted(ar.iteritems()) ):
            self.assertEqual( items[i], k )

    def test1(self):
        """
        Simple tests: elementary and intermediate
        uncertain numbers restored OK? Saving
        all objects.
        
        """
        context._context = Context()
        
        x = ureal(1,1)
        y = ureal(2,1)
        z = result( x + y )

        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)
        ar._freeze()

        # Unpack and check 
        context._context = Context()
        ar._thaw()

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual(x.x,x1.x)
        self.assertEqual(x.u,x1.u)
        self.assertEqual(x.df,x1.df)
        self.assertEqual(y.x,y1.x)
        self.assertEqual(y.u,y1.u)
        self.assertEqual(y.df,y1.df)
        self.assertEqual(z.x,z1.x)
        self.assertEqual(z.u,z1.u)
        self.assertEqual(z.df,z1.df)

        a = component(z,x)
        b = component(z1,x1)
        self.assertTrue( equivalent(a,b,TOL) )

    def test2(self):
        """
        Logical correlation works - I
        with all objects archived
        
        """
        context._context = Context()
        
        x1 = ureal(1,1)
        x2 = ureal(1,1)
        x3 = ureal(1,1)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack and test
        context._context = Context()
        ar._thaw()
        
        y1, y2, y3 = ar.extract('x1','x2','x3')

        self.assertEqual( str(x1),str(y1) )
        self.assertEqual( str(x2),str(y2) )
        self.assertEqual( str(x3),str(y3) )

        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = component(x6,x4)
        b = component(y6,y4)
        self.assertTrue( equivalent(a,b,TOL) )

        a = component(x6,x2)
        b = component(y6,y2)
        self.assertTrue( equivalent(a,b,TOL) )

        a = get_correlation(x4,x5)
        b = get_correlation(y4,y5)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test3(self):
        """
        Logical correlation works - II
        with all objects archived but not all restored
        
        """
        context._context = Context()

        x1 = ureal(1,1)
        x2 = ureal(1,1)
        x3 = ureal(1,1)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = component(x6,x4)
        b = component(y6,y4)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test3b(self):
        """
        Logical correlation works - III
        with not all objects archived or restored
        
        """
        context._context = Context()

        x1 = ureal(1,1)
        x2 = ureal(1,1)
        x3 = ureal(1,1)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = component(x6,x4)
        b = component(y6,y4)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test4(self):
        """
        Explicit correlation works - I
        with all objects archived
        
        """
        context._context = Context()

        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,independent=False)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        set_correlation(0.5,x1,x2)
        set_correlation(-0.25,x3,x2)

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        y1, y2, y3 = ar.extract('x1','x2','x3')

        a = get_correlation(x1,x2)
        b = get_correlation(y1,y2)
        self.assertTrue( equivalent(a,b,TOL) )

        a = get_correlation(x3,x2)
        b = get_correlation(y3,y2)
        self.assertTrue( equivalent(a,b,TOL) )

        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = get_correlation(x4,x5)
        b = get_correlation(y4,y5)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test5(self):
        """
        Explicit correlation works - II
        with only some objects restored
        
        """
        context._context = Context()

        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,independent=False)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        set_correlation(0.5,x1,x2)
        set_correlation(-0.25,x3,x2)

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = get_correlation(x4,x5)
        b = get_correlation(y4,y5)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test5b(self):
        """
        Explicit correlation works - II
        with only some objects archived and restored
        
        """
        context._context = Context()

        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,independent=False)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        set_correlation(0.5,x1,x2)
        set_correlation(-0.25,x3,x2)

        ar = persistence.Archive()
        ar.add(x4=x4,x5=x5,x6=x6)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()
        y4, y5, y6 = ar.extract('x4','x5','x6')
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = get_correlation(x4,x5)
        b = get_correlation(y4,y5)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def test6(self):
        """
        Same as test 5 but with [] indexing
        
        """
        context._context = Context()
        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,independent=False)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        set_correlation(0.5,x1,x2)
        set_correlation(-0.25,x3,x2)

        ar = persistence.Archive()
        ar['x1'] = x1
        ar['x2'] = x2
        ar['x3'] = x3
        ar['x4'] = x4
        ar['x5'] = x5
        ar['x6'] = x6
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        y4 = ar['x4']
        y5 = ar['x5']
        y6 = ar['x6']
        
        self.assertEqual( str(x4),str(y4) )
        self.assertEqual( str(x5),str(y5) )
        self.assertEqual( str(x6),str(y6) )

        a = get_correlation(x4,x5)
        b = get_correlation(y4,y5)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y6._u_components) )
        self.assertTrue( is_ordered(y6._d_components) )
        self.assertTrue( is_ordered(y6._i_components) )

    def testGUMH2(self):
        """
        GUM H2 as an ensemble
        
        """
        # NB this is a dodgy test, because we are reusing 
        # the same archive object in a way that is not intended.
        
        context._context = Context()
        
        x = [4.999,0.019661,1.04446]
        u = [0.0032,0.0000095,0.00075]
        
        v,i,phi = multiple_ureal(x,u,5)
        
        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)
        
        r = v * cos(phi)/i
        x = v * sin(phi)/i
        z = v/i

        ar = persistence.Archive()
        ar.add(v=v,i=i,phi=phi)
        ar._freeze()
                
        # unpack without elementary UNs and test

        # This is a fudge. Normally, a different 
        # archive object would be created from a 
        # file or string. 
        context._context = Context()
        ar._thaw()
        
        v1, i1, phi1 = ar.extract('v','i','phi')
        
        # self.assertTrue( v1._node in c2._ensemble )
        # self.assertTrue( i1._node in c2._ensemble )
        # self.assertTrue( phi1._node in c2._ensemble )

        self.assertTrue( hasattr(v1._node, 'ensemble') )
        self.assertTrue( v1._node.uid in v1._node.ensemble )
        self.assertTrue( hasattr(i1._node, 'ensemble') )
        self.assertTrue( i1._node.uid in i1._node.ensemble )
        self.assertTrue( hasattr(phi1._node, 'ensemble') )
        self.assertTrue( phi1._node.uid in phi1._node.ensemble )

        
        self.assertEqual( repr(v1), repr(v) )
        self.assertEqual( repr(i1) , repr(i) )
        self.assertEqual( repr(phi1) , repr(phi) )
        
        self.assertTrue( equivalent( get_correlation(v1,i1),-0.36) )
        self.assertTrue( equivalent( get_correlation(v1,phi1),0.86) )
        self.assertTrue( equivalent( get_correlation(i1,phi1),-0.65) )
           
        r1 = v1 * cos(phi1)/ i1
        x1 = v1 * sin(phi1)/ i1
        z1 = v1 / i1

        # The DoF calculation would fail if the inputs
        # are not part of the same ensemble.    
        self.assertTrue( equivalent( dof(r1),5,TOL) )
        self.assertTrue( equivalent( dof(x1),5,TOL) )
        self.assertTrue( equivalent( dof(z1),5,TOL) )

        self.assertTrue( equivalent( value(r1),value(r),TOL) )
        self.assertTrue( equivalent( value(x1),value(x),TOL) )
        self.assertTrue( equivalent( value(z1),value(z),TOL) )

        self.assertTrue( equivalent( uncertainty(r1),uncertainty(r),TOL) )
        self.assertTrue( equivalent( uncertainty(x1),uncertainty(x),TOL) )
        self.assertTrue( equivalent( uncertainty(z1),uncertainty(z),TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(r1._u_components) )
        self.assertTrue( is_ordered(x1._u_components) )
        self.assertTrue( is_ordered(z1._u_components) )

    def testGUMH2_b(self):
        """
        GUM H2 as an ensemble. Only save the results this time.
        
        """
        context._context = Context()

        x = [4.999,0.019661,1.04446]
        u = [0.0032,0.0000095,0.00075]
        
        v,i,phi = multiple_ureal(x,u,5)
        
        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = result( v * cos(phi)/ i )
        x = result( v * sin(phi)/ i )
        z = result( v / i )

        ar = persistence.Archive()
        ar.add(r=r,x=x,z=z)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()

        r1, x1, z1 = ar.extract('r','x','z')
        
        self.assertEqual( repr(r1), repr(r) )
        self.assertEqual( repr(x1) , repr(x) )
        self.assertEqual( repr(z1) , repr(z) )
               
        # The DoF calculation would fail if the inputs
        # are not part of the same ensemble.    
        self.assertTrue( equivalent( dof(r1),5,TOL) )
        self.assertTrue( equivalent( dof(x1),5,TOL) )
        self.assertTrue( equivalent( dof(z1),5,TOL) )

        self.assertTrue( equivalent( value(r1),value(r),TOL) )
        self.assertTrue( equivalent( value(x1),value(x),TOL) )
        self.assertTrue( equivalent( value(z1),value(z),TOL) )

        self.assertTrue( equivalent( uncertainty(r1),uncertainty(r),TOL) )
        self.assertTrue( equivalent( uncertainty(x1),uncertainty(x),TOL) )
        self.assertTrue( equivalent( uncertainty(z1),uncertainty(z),TOL) )
        
        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(r1._u_components) )
        self.assertTrue( is_ordered(x1._u_components) )
        self.assertTrue( is_ordered(z1._u_components) )

    def test_intermediates_multipath(self):
        """
        Check the sensitivities when the tree structure
        has intermediates that depend on intermediates

        In this case y is directly connected to x1, x4 and x5
        but x4 and x5 also depend on x1 and x5 depends on x4

        """
        context._context = Context()
        
        x1 = ureal(1.2,0.6,label='x1')
        x2 = ureal(2.5,1.6,label='x2')
        x3 = ureal(-5.3,.77,label='x3')
        x4 = result( 2.0 * (x1 + x2) )
        x5 = result( (x4 + x3)**2 )
        y = result( sqrt(x5 + x4 + x1) )

        ar = persistence.Archive()
        ar.add(y=y,x4=x4,x5=x5,x1=x1)
        ar._freeze()

        # unpack without elementary uns and test
        context._context = Context()
        ar._thaw()
        yy, xx4, xx5, xx1 = ar.extract('y','x4','x5','x1')

        self.assertTrue(
            equivalent(
                rp.u_component(y,x4),
                rp.u_component(yy,xx4),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(y,x5),
                rp.u_component(yy,xx5),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(y,x1),
                rp.u_component(yy,xx1),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(x5,x1),
                rp.u_component(xx5,xx1),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(x4,x1),
                rp.u_component(xx4,xx1),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(x5,x4),
                rp.u_component(xx5,xx4),
                TOL
            )
        )

        self.assertTrue(
            equivalent(
                rp.u_component(x4,x5),
                rp.u_component(xx4,xx5),
                TOL
            )
        )
     
    def test_complex_1(self):
        """
        Simple arithmetic - all objects stored
        and restored.
        
        """
        context._context = Context()
        
        x = ucomplex(1,[1,2],4)
        y = ucomplex(1,[3,12],3)
        z = result( x * y )

        ar = persistence.Archive()
        ar.add(x=x)
        ar.add(y=y)
        ar.add(z=z)
        ar._freeze()

        # unpack 
        context._context = Context()
        ar._thaw()

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(x1), repr(x) )

        self.assertTrue( equivalent_complex(x1.x,x.x) )
        self.assertTrue( equivalent_sequence(x1.u,x.u) )
        self.assertTrue( equivalent(x1.df,x.df) )

        self.assertTrue( equivalent_complex(y1.x,y.x) )
        self.assertTrue( equivalent_sequence(y1.u,y.u) )
        self.assertTrue( equivalent(y1.df,y.df) )
        
        self.assertTrue( equivalent_complex(z1.x,z.x) )
        self.assertTrue( equivalent_sequence(z1.u,z.u) )
        self.assertTrue( equivalent(z1.df,z.df) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(y1.real._u_components) )
        self.assertTrue( is_ordered(x1.real._u_components) )
        self.assertTrue( is_ordered(z1.real._u_components) )
        self.assertTrue( is_ordered(y1.imag._u_components) )
        self.assertTrue( is_ordered(x1.imag._u_components) )
        self.assertTrue( is_ordered(z1.imag._u_components) )

    def test_complex_2(self):
        """
        Complex logical correlation
        
        """
        context._context = Context()
        
        x1 = ucomplex(1,[3,2],4)
        x2 = ucomplex(1,[1,1],5)
        x3 = ucomplex(1,[4,5],6)

        x4 = result( x1*x2 )
        x5 = result( x2*x3 )

        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(x4=x4,x5=x5,x6=x6)
        ar._freeze()

        context._context = Context()
        ar._thaw()
        y4, y5, y6 = ar.extract('x4','x5','x6')

        self.assertEqual( repr(x4), repr(y4) )
        self.assertEqual( repr(x5), repr(y5) )
        self.assertEqual( repr(x6), repr(y6) )

        self.assertTrue( equivalent_complex(x4.x,y4.x) )
        self.assertTrue( equivalent_sequence(x4.u,y4.u) )
        self.assertTrue( equivalent(x4.df,y4.df) )

        self.assertTrue( equivalent_complex(x5.x,y5.x) )
        self.assertTrue( equivalent_sequence(x5.u,y5.u) )
        self.assertTrue( equivalent(x5.df,y5.df) )
        
        self.assertTrue( equivalent_complex(x6.x,y6.x) )
        self.assertTrue( equivalent_sequence(x6.u,y6.u) )
        self.assertTrue( equivalent(x6.df,y6.df) )

        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x5),
            get_correlation(y6,y5)
        ) )
                               
        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x4),
            get_correlation(y6,y4)
        ) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(x6.real._u_components) )
        self.assertTrue( is_ordered(x6.imag._u_components) )

    def test_complex_3(self):
        """
        Complex logical and external correlation
        
        """
        context._context = Context()
        
        x1 = ucomplex(1,[3,2],independent=False)
        x2 = ucomplex(1,[1,1],independent=False)
        x3 = ucomplex(1,[4,5])

        R = [.1,.2,.3,.4]
        set_correlation(R,x1,x2)
        # Make sure we get what we expect
        self.assertTrue( equivalent_sequence(
            R, get_correlation(x1,x2)
        ))
        
        x4 = result( x1*x2 )
        x5 = result( x2*x3 )

        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6)
        ar._freeze()

        context._context = Context()
        ar._thaw( )
        y4, y5, y6 = ar.extract('x4','x5','x6')

        self.assertEqual( repr(x4), repr(y4) )
        self.assertEqual( repr(x5), repr(y5) )
        self.assertEqual( repr(x6), repr(y6) )
        
        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x5),
            get_correlation(y6,y5)
        ) )
                               
        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x4),
            get_correlation(y6,y4)
        ) )

        self.assertTrue( equivalent_sequence(
            variance(x6),
            variance(y6)
        ))
        
        # Now look at the elementary influences
        y1, y2 = ar.extract('x1','x2')
        self.assertTrue( equivalent_sequence(
            R, get_correlation(y1,y2)
        ))

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(x6.real._u_components) )
        self.assertTrue( is_ordered(x6.imag._u_components) )

    def test_complex_4(self):
        """
        Same as 'complex_3' but using [] indexing
        
        """
        context._context = Context()

        x1 = ucomplex(1,[3,2],independent=False)
        x2 = ucomplex(1,[1,1],independent=False)
        x3 = ucomplex(1,[4,5])

        R = [.1,.2,.3,.4]
        set_correlation(R,x1,x2)
        # Make sure we get what we expect
        self.assertTrue( equivalent_sequence(
            R, get_correlation(x1,x2)
        ))
        
        x4 = result( x1*x2 )
        x5 = result( x2*x3 )

        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar['x1'] = x1 
        ar['x2'] = x2 
        ar['x3'] = x3 
        ar['x4'] = x4 
        ar['x5'] = x5 
        ar['x6'] = x6 
        ar._freeze()

        context._context = Context()
        ar._thaw( )
        y4 = ar['x4']
        y5 = ar['x5']
        y6 = ar['x6']

        self.assertEqual( repr(x4), repr(y4) )
        self.assertEqual( repr(x5), repr(y5) )
        self.assertEqual( repr(x6), repr(y6) )
        
        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x5),
            get_correlation(y6,y5)
        ) )
                               
        self.assertTrue( equivalent_sequence(
            get_correlation(x6,x4),
            get_correlation(y6,y4)
        ) )

        self.assertTrue( equivalent_sequence(
            variance(x6),
            variance(y6)
        ))
        
        # Now look at the elementary influences
        y1, y2 = ar.extract('x1','x2')
        self.assertTrue( equivalent_sequence(
            R, get_correlation(y1,y2)
        ))

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(x6.real._u_components) )
        self.assertTrue( is_ordered(x6.imag._u_components) )

    def test_complex_5(self):
        """
        Archive intermediate complex components
        
        """
        context._context = Context()

        x1 = ucomplex(1,[3,2],independent=False)
        x2 = ucomplex(1,[1,1],independent=False)
        x3 = ucomplex(1,[4,5])

        R = [.1,.2,.3,.4]
        set_correlation(R,x1,x2)
        # Make sure we get what we expect
        self.assertTrue( equivalent_sequence(
            R, get_correlation(x1,x2)
        ))
        
        x4 = result( x1*x2 )
        x5 = result( x2*x3 )

        x6 = result( x4 + x5  )
        x7 = result( mag_squared(x6) + mag_squared(x4) )

        ar = persistence.Archive()
        ar['x1'] = x1 
        ar['x2'] = x2 
        ar['x3'] = x3 
        ar['x4'] = x4.real  # just the real component
        ar['x5'] = x5 
        ar['x6'] = x6
        ar['x7'] = x7
        ar._freeze()

        context._context = Context()
        ar._thaw()
        y4 = ar['x4']
        y5 = ar['x5']
        y6 = ar['x6']
        y7 = ar['x7']

        self.assertEqual( repr(x4.real), repr(y4) )
        self.assertEqual( repr(x5), repr(y5) )
        self.assertEqual( repr(x6), repr(y6) )
        
        # The restored complex should have a link between the component nodes
        self.assertEqual( x5.real._node.complex, y5.real._node.complex )
        self.assertEqual( y5.imag._node.complex, y5.real._node.complex )
        
        # Archived as a complex influence of a complex quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x6,x5),
            rp.u_component(y6,y5)
        ) )

        # Archived as a real influence of a complex quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x6,x4.real),
            rp.u_component(y6,y4)
        ) )

        # Archived as a complex influence of a real quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x7,x5),
            rp.u_component(y7,y5)
        ))
        
        # Archived as a real influence of a real quantity                
        self.assertTrue( equivalent(
            rp.u_component(x7,x4.real),
            rp.u_component(y7,y4)
        ))

    def test_nop_intermediates(self):
        """
        Archiving irrelevant intermediate components
        
        """
        context._context = Context()

        x1 = ucomplex(1,[3,2])
        x2 = ureal(1,1)
        x3 = constant(1.3)
        
        x4 = result( x1*x2 )
        x5 = result( x2*x3 )

        x6 = result( x4 + x5 )
        x7 = result( mag_squared(x6) )

        ar = persistence.Archive()
        ar.add(
            x1=x1,x2=x2,x3=x3,x4=x4.real,
            x5=x5,x6=x6
        )
        ar.add(x7=x7)
        ar._freeze()

        context._context = Context()
        ar._thaw( )
        y1 = ar['x1']
        y2 = ar['x2']
        y3 = ar['x3']
        y6 = ar['x6']
        y7 = result( ar['x7'] )

        self.assertEqual( repr(x1), repr(y1) )
        self.assertEqual( repr(x2), repr(y2) )
        self.assertEqual( repr(x3), repr(y3) )
        self.assertEqual( repr(x6), repr(y6) )
        self.assertEqual( repr(x7), repr(y7) )

        # Archived as a complex influence of a complex quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x6,x1),
            rp.u_component(y6,y1)
        ) )

        # Archived as a real influence of a complex quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x6,x2),
            rp.u_component(y6,y2)
        ) )

        # Archived as a complex influence of a real quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x7,x1),
            rp.u_component(y7,y1)
        ))
        
        # Archived as a real influence of a real quantity                
        self.assertTrue( equivalent(
            rp.u_component(x7,x2),
            rp.u_component(y7,y2)
        ))

        # Archived as a constant influence of a real quantity 
        self.assertTrue( equivalent(
            rp.u_component(x7,x3),
            rp.u_component(y7,y3)
        ))

        # Archived as a constant influence of a complex quantity                
        self.assertTrue( equivalent_sequence(
            rp.u_component(x6,x3),
            rp.u_component(y6,y3)
        ) )
                         
    def test_with_file(self):
        """
        Save to a file and then restore by reading
        """
        wdir = os.getcwd()
        fname = 'test_file.pkl'
        path = os.path.join(wdir,fname)

        context._context = Context()
        x = ureal(1,1)
        y = ureal(2,1)
        z = result( x + y )

        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        f = open(path,'wb')
        persistence.dump(f,ar)
        f.close()

        context._context = Context()
        f = open(path,'rb')
        ar = persistence.load(f)
        f.close()
        os.remove(path)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )

    def test_with_file2(self):
        """
        Save to a file and then restore several times
        to test the effectiveness of GTC's uid system.
        
        """
        wdir = os.getcwd()
        fname = 'test_file.pkl'
        path = os.path.join(wdir,fname)

        context._context = Context()
        
        x = ureal(1,1,3,label='x')
        y = ureal(2,1,4,label='y')
        z = result( x + y )

        ar = persistence.Archive()

        # Saving only `z` means that when the archive
        # is restored `x` and `y` are not recreated as UNs.
        # However, Leaf nodes are created. We need to make sure 
        # that only one Leaf node gets created.
        
        ar.add(z=z)

        with open(path,'wb') as f:
            persistence.dump(f,ar)

        context._context = Context()
        
        with open(path,'rb') as f:
            ar1 = persistence.load(f)

        z1 = ar1.extract('z')

        with open(path,'rb') as f:
            # The attempt to create the node again is caught
            self.assertRaises(RuntimeError,persistence.load,f)

        os.remove(path)
 
    def test_with_file3(self):
        """
        Dependent elementary UNs
        """
        wdir = os.getcwd()
        fname = 'test_file.pkl'
        path = os.path.join(wdir,fname)

        context._context = Context()

        x = ureal(1,1,independent=False)
        y = ureal(2,1,independent=False)
        
        r = 0.5
        set_correlation(r,x,y)
        
        z = result( x + y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        with open(path,'wb') as f:
            persistence.dump(f,ar)

        context._context = Context()
        with open(path,'rb') as f:
            ar = persistence.load(f)
            
        os.remove(path)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x,y), r )

    def test_with_file4(self):
        """
        Correlations with finite DoF
        """
        wdir = os.getcwd()
        fname = 'test_file.pkl'
        path = os.path.join(wdir,fname)

        context._context = Context()

        x,y = multiple_ureal([1,2],[1,1],4)
        
        r = 0.5
        set_correlation(r,x,y)
        
        z = result( x + y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        with open(path,'wb') as f:
            persistence.dump(f,ar)

        context._context = Context()
        with open(path,'rb') as f:
            ar = persistence.load(f)

        os.remove(path)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x,y), r )
 
    def test_with_file5(self):
        """
        Complex
        """
        wdir = os.getcwd()
        fname = 'test_file.pkl'
        path = os.path.join(wdir,fname)

        context._context = Context()

        x = ucomplex(1,[10,2,2,10],5)
        y = ucomplex(1-6j,[10,2,2,10],7)
        
        z = result( log( x * y ) )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        with open(path,'wb') as f:
            persistence.dump(f,ar)

        context._context = Context()
        with open(path,'rb') as f:
            ar = persistence.load(f)

        os.remove(path)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
 
    def test_with_file6(self):
        """
        The management of new leaf nodes must be 
        able to recognise when an identical leaf 
        has been declared in context already.
        
        """
        wdir = os.getcwd()
        fname1 = 'test_file_1.pkl'
        fname2 = 'test_file_2.pkl'
        path1 = os.path.join(wdir,fname1)
        path2 = os.path.join(wdir,fname2)

        context._context = Context()
        
        x = ureal(1,1,3,label='x')
        y = ureal(2,1,4,label='y')
        z1 = result( x + y )
        z2 = result( 2*x - 3*y )
        
        ar1 = persistence.Archive()
        ar2 = persistence.Archive()

        # Saving only `z` means that when the archive
        # is restored Leaf nodes are created. 
        # We need to make sure that only one Leaf node 
        # gets created.
        
        ar1.add(z1=z1)
        with open(path1,'wb') as f:
            persistence.dump(f,ar1)

        ar2.add(z2=z2)
        with open(path2,'wb') as f:
            persistence.dump(f,ar2)

        # Now both archives have the same 
        # Leaf node dependencies, restoring 
        # the first one re-creates the leaves
        # so the second one should quietly 
        # recognise that the leaves are the same. 
        context._context = Context()
        
        with open(path1,'rb') as f:
            ar1 = persistence.load(f)

        z1 = ar1.extract('z1')

        with open(path2,'rb') as f:
            ar2 = persistence.load(f)

        z2 = ar2.extract('z2')

        os.remove(path1)
        os.remove(path2)
 
    def test_with_string(self):
        """
        Simple save with intermediate 
        """
        context._context = Context()

        x = ureal(1,1)
        y = ureal(2,1)
        z = result( x + y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        db = persistence.dumps(ar)

        context._context = Context()
        ar = persistence.loads(db)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )

    def test_with_string2(self):
        """
        Save to a file and then restore several times
        to test the effectiveness of GTC's uid system.
        
        """
        context._context = Context()
        
        x = ureal(1,1,3,label='x')
        y = ureal(2,1,4,label='y')
        z = result( x + y )

        ar = persistence.Archive()

        # Saving only `z` means that when the archive
        # is restored `x` and `y` are not recreated as UNs.
        # However, Leaf nodes are created. We need to make sure 
        # that only one Leaf node gets created.
        
        ar.add(z=z)

        s = persistence.dumps(ar)

        context._context = Context()       
        ar1 = persistence.loads(s)

        z1 = ar1.extract('z')

        # The attempt to create the node again is caught
        self.assertRaises(RuntimeError,persistence.loads,s)
 
    def test_with_string3(self):
        """
        Dependent elementary UNs
        """
        context._context = Context()

        x = ureal(1,1,independent=False)
        y = ureal(2,1,independent=False)
        
        r = 0.5
        set_correlation(r,x,y)
        
        z = result( x + y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        db = persistence.dumps(ar)

        context._context = Context()
        ar = persistence.loads(db)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x,y), r )
 
    def test_with_string4(self):
        """
        Correlations with finite DoF
        """
        context._context = Context()

        x,y = multiple_ureal([1,2],[1,1],4)
        
        r = 0.5
        set_correlation(r,x,y)
        
        z = result( x + y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        db = persistence.dumps(ar)

        context._context = Context()
        ar = persistence.loads(db)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x,y), r )
 
    def test_with_string5(self):
        """
        Complex
        """
        context._context = Context()

        x = ucomplex(1,[10,2,2,10],5)
        y = ucomplex(1-6j,[10,2,2,10],7)
        
        z = result( x * y )
        
        ar = persistence.Archive()

        ar.add(x=x,y=y,z=z)

        db = persistence.dumps(ar)

        context._context = Context()
        ar = persistence.loads(db)

        x1, y1, z1 = ar.extract('x','y','z')

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
    def test_multiple_names(self):
        """
        Different names may apply to the same UN
        
        """
        context._context = Context()

        x1 = ureal(1,1)
        x2 = ureal(1,1)
        x3 = ureal(1,1)
        x4 = result( x1 + x2 )
        x5 = result( x2 + x3 )
        x6 = result( x4 + x5 )

        ar = persistence.Archive()
        ar.add(
            x1=x1,x2=x2,x3=x3,x4=x4,x5=x5,x6=x6,
            z1=x1,z4=x4,z6=x6
        )
        ar._freeze()

        # unpack and test
        context._context = Context()
        ar._thaw()
        y1, y2, y3 = ar.extract('x1','x2','x3')
        yz1 = ar['z1']
        
        # This used to be the case, but thawing now allows multiple UNs
        # self.assertTrue( yz1 is y1 )
        
        self.assertTrue( yz1._node.uid is y1._node.uid )
        self.assertEqual( repr(x1), repr(y1) )
        self.assertEqual( repr(yz1),repr(y1) )

        # NB most of these tests are redundant
        # now that multiple names are restored
        # to the same object. 
        y4, y5, y6 = ar.extract('x4','x5','x6')
        yz4 = ar['z4']
        yz6 = ar['z6']
        # self.assertTrue( yz4 is y4 )
        # self.assertTrue( yz6 is y6 )
        self.assertTrue( yz4._node.uid is y4._node.uid )
        self.assertTrue( yz6._node.uid is y6._node.uid )
        
        self.assertEqual( repr(x4),repr(y4) )
        self.assertEqual( repr(yz4), repr(y4) )
        self.assertEqual( repr(x6), repr(y6) )
        self.assertEqual( repr(yz6), repr(y6) )

        a = component(x6,x4)
        b = component(y6,y4)
        self.assertTrue( equivalent(a,b,TOL) )

        b = component(yz6,y4)
        self.assertTrue( equivalent(a,b,TOL) )

        b = component(yz6,yz4)
        self.assertTrue( equivalent(a,b,TOL) )

        # Make sure the vector of components is well-formed
        self.assertTrue( is_ordered(yz6._u_components) )
        
    def test_archive_with_previously_archived(self):
        """
        Make sure that an archive can be created that
        includes some UNs that have been restored from
        another archive.

        NB some tests here imposed interchangeability of
        objects that belong to different contexts. But
        this is not a requirement, so I have commented
        out. We should never have to work with a mixture 
        of uncertain number instances created by different
        Contexts.
        
        """
        # Archive some simple stuff
        # This is the first archive. 
        context._context = Context()

        x1 = ureal(1,1,5,label='x1')
        x2 = ucomplex(1,1,7,label='x2')
        x3 = result( x1 + x2 )
        self.assertTrue( x3.real.is_intermediate )

        ar = persistence.Archive()
        ar.add(x1=x1,x2=x2,x3=x3)
        ar._freeze()

        # Now add some more stuff and store
        # in a second archive
        context._context = Context()
        ar._thaw()
        
        y1 = ar['x1']
        y2 = ar['x2']
        y3 = ar['x3'] 

        self.assertTrue( y1.is_elementary )
        self.assertTrue( y2.is_elementary )
        self.assertTrue( not y3.is_elementary )

        self.assertEqual(y3.real._node.uid,x3.real._node.uid)
        self.assertTrue( y3.real.is_intermediate )

        self.assertTrue(
            equivalent(component(x3,x1),component(y3,y1))
        )

        for i,j in zip(
            x3.real._i_components._index,
            y3.real._i_components._index
        ):
            self.assertEqual(i.uid,j.uid)
            self.assertTrue( equivalent(
                x3.real._i_components[i], 
                y3.real._i_components[j],
                TOL
            ))

        y4 = ureal(10,1,23)
        y5 = ucomplex(2.5+7j,2,19)
        y6 = result( y3 * (y4 + y5) )
        
        ar = persistence.Archive()
        ar.add(
            y1=y1,
            y2=y2,
            y3=y3,
            y4=y4,y5=y5,y6=y6
        )
        ar._freeze()

        # Now restore into a third Context
        context._context = Context()
        ar._thaw()
        
        z1 = ar['y1']
        z2 = ar['y2']
        z3 = ar['y3']

        self.assertTrue( z1.is_elementary )
        self.assertTrue( z2.is_elementary )
        self.assertTrue( not z3.is_elementary )

        # First just check that we get the same objects
        # and that the values are correct
        self.assertEqual(z1._node.uid,y1._node.uid)
        self.assertEqual(z1._node.uid[1],x1._node.uid[1])
        self.assertEqual(z1._node.uid[0],x1._node.uid[0])

        self.assertTrue(
            equivalent(value(z1),value(y1))
        )
        self.assertTrue(
            equivalent(value(z1),value(x1))
        )
        self.assertTrue(
            equivalent(uncertainty(z1),uncertainty(y1))
        )
        self.assertTrue(
            equivalent(uncertainty(z1),uncertainty(x1))
        )
        self.assertTrue(
            equivalent(dof(z1),dof(y1))
        )
        self.assertTrue(
            equivalent(dof(z1),dof(x1))
        )

        self.assertEqual(z2.imag._node.uid,y2.imag._node.uid)
        self.assertEqual(z2.real._node.uid,y2.real._node.uid)
        self.assertEqual(z2.real._node.uid,x2.real._node.uid)
        self.assertEqual(z2.imag._node.uid,x2.imag._node.uid)
        self.assertTrue(
            equivalent(value(z2.real),value(y2.real))
        )
        self.assertTrue(
            equivalent(value(z2.real),value(x2.real))
        )
        self.assertTrue(
            equivalent(uncertainty(z2.real),uncertainty(y2.real))
        )
        self.assertTrue(
            equivalent(uncertainty(z2.real),uncertainty(x2.real))
        )
        self.assertTrue(
            equivalent(dof(z2),dof(y2))
        )
        self.assertTrue(
            equivalent(dof(z2),dof(x2))
        )
        self.assertEqual(z2.imag._node.uid,y2.imag._node.uid)
        self.assertEqual(z2.imag._node.uid,x2.imag._node.uid)
        
        self.assertTrue(
            equivalent(value(z2.imag),value(y2.imag))
        )
        self.assertTrue(
            equivalent(value(z2.imag),value(x2.imag))
        )
        self.assertTrue(
            equivalent(uncertainty(z2.imag),uncertainty(y2.imag))
        )
        self.assertTrue(
            equivalent(uncertainty(z2.imag),uncertainty(x2.imag))
        )

        self.assertTrue(
            equivalent(value(z3.imag),value(y3.imag))
        )
        self.assertTrue(
            equivalent(value(z3.imag),value(x3.imag))
        )
        self.assertTrue(
            equivalent(uncertainty(z3.imag),uncertainty(y3.imag))
        )
        self.assertTrue(
            equivalent(uncertainty(z3.imag),uncertainty(x3.imag))
        )

        # Uncertainty components
        self.assertTrue(
            equivalent(component(x3,x1),component(z3,z1))
        )
        
        self.assertTrue(
            equivalent_sequence(rp.u_component(x3,x2),rp.u_component(z3,z2))
        )

        # Continue to restore
        z4 = ar['y4']
        z5 = ar['y5']
        z6 = ar['y6']
        
        # Make sure the indexing of components is still ordered
        self.assertTrue( is_ordered(z6.real._u_components) )

        # First just check that we get the same objects
        # and that the values are correct
        self.assertEqual(z4._node.uid,y4._node.uid)
        self.assertTrue(
            equivalent(value(z4),value(y4))
        )
        self.assertTrue(
            equivalent(uncertainty(z4),uncertainty(y4))
        )
        self.assertTrue(
            equivalent(dof(z4),dof(y4))
        )

        self.assertEqual(z5.real._node.uid,y5.real._node.uid)
        self.assertTrue(
            equivalent(value(z5.real),value(y5.real))
        )
        self.assertTrue(
            equivalent(uncertainty(z5.real),uncertainty(y5.real))
        )
        self.assertTrue(
            equivalent(uncertainty(z5.imag),uncertainty(y5.imag))
        )

        # Components
        # We will get components wrt elementary inputs
        self.assertTrue(
            equivalent_sequence(rp.u_component(z6,z2),rp.u_component(y6,y2))
        )
        # And intermediates (provided this was requested)
        self.assertTrue(
            equivalent_sequence(rp.u_component(z6,z3),rp.u_component(y6,y3))
        )

        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


from collections import namedtuple

VarianceAndDof = namedtuple('VarianceAndDof','cv, df')
""":obj:`~collections.namedtuple`: Values of the variance and degrees of freedom.
 
.. attribute:: cv
  
   Variance.
  
.. attribute:: df
  
   :class:`float`: Degrees of freedom.
 
"""

VarianceCovariance = namedtuple('VarianceCovariance','rr, ri, ir, ii')
""":obj:`~collections.namedtuple`: Values of variance-covariance for a complex quantity
 
.. attribute:: rr
  
   :class:`float`: variance in the real component
  
.. attribute:: ri
  
   :class:`float`: covariance between th real and imaginary components
 
.. attribute:: ir
  
   :class:`float`: covariance between th real and imaginary components
   
.. attribute:: ii
  
   :class:`float`:  variance in the imaginary component

   """

StandardUncertainty = namedtuple('StandardUncertainty','real,imag')
""":obj:`~collections.namedtuple`: Standard uncertainty values of a complex quantity
 
.. attribute:: real
  
   :class:`float`: standard uncertainty in the real component
   
.. attribute:: imag
  
   :class:`float`: standard uncertainty in the imaginary component
   
"""

StandardDeviation = namedtuple('StandardDeviation','real,imag')
""":obj:`~collections.namedtuple`: Standard deviation values of a complex quantity
 
.. attribute:: real
  
   :class:`float`: standard deviation in the real component
   
.. attribute:: imag
  
   :class:`float`: standard deviation in the imaginary component
   
"""

ComponentOfUncertainty = namedtuple('ComponentOfUncertainty','rr, ri, ir, ii')
""":obj:`~collections.namedtuple`: Component of uncertainty values for a complex quantity
 
.. attribute:: rr
  
   :class:`float`: real component with respect to real component
  
.. attribute:: ri
  
   :class:`float`: real component with respect to imaginary component
 
.. attribute:: ir
  
   :class:`float`: imaginary component with respect to real component
   
.. attribute:: ii
  
   :class:`float`: imaginary component with respect to imaginary component

   """

JacobianMatrix = namedtuple('JacobianMatrix','rr, ri, ir, ii')
""":obj:`~collections.namedtuple`: Partial derivatives of one complex with respect to another
 
.. attribute:: rr
  
   :class:`float`: real component with respect to real component
  
.. attribute:: ri
  
   :class:`float`: real component with respect to imaginary component
 
.. attribute:: ir
  
   :class:`float`: imaginary component with respect to real component
   
.. attribute:: ii
  
   :class:`float`: imaginary component with respect to imaginary component

   """

Influence = namedtuple('Influence','label, u')
""":obj:`~collections.namedtuple`: label and value of a component of uncertainty
 
.. attribute:: label
  
   :class:`str`: influence quantity label
   
.. attribute:: u
  
   :class:`float`: component of uncertainty due to influence quantity
   
"""
CorrelationMatrix = namedtuple("CorrelationMatrix","rr,ri,ir,ii")
""":obj:`~collections.namedtuple`: Correlation coefficients for a pair of quantities ``x`` and ``y``
 
.. attribute:: rr
  
   :class:`float`: correlation between ``x.real`` and ``y.real``
  
.. attribute:: ri
  
   :class:`float`: correlation between ``x.real`` and ``y.imag``
 
.. attribute:: ir
  
   :class:`float`: correlation between ``x.imag`` and ``y.real``
   
.. attribute:: ii
  
   :class:`float`: correlation between ``x.imag`` and ``y.imag``

   """

CovarianceMatrix = namedtuple("CovarianceMatrix","rr,ri,ir,ii")
""":obj:`~collections.namedtuple`: Values of covariance for a pair of quantities ``x`` and ``y``
 
.. attribute:: rr
  
   :class:`float`: covariance between ``x.real`` and ``y.real``
  
.. attribute:: ri
  
   :class:`float`: covariance between ``x.real`` and ``y.imag``
 
.. attribute:: ir
  
   :class:`float`: covariance between ``x.imag`` and ``y.real``
   
.. attribute:: ii
  
   :class:`float`: covariance between ``x.imag`` and ``y.imag``

   """

InterceptSlope = namedtuple('InterceptSlope','a b')
""":obj:`~collections.namedtuple`: Values for intercept ``a`` and slope ``b``
 
.. attribute:: a
  
   :class:`~.lib.UncertainReal`: intercept
  
.. attribute:: b
  
   :class:`~.lib.UncertainReal`: slope
 """

GroomedUncertainReal = namedtuple('ureal','x u df label precision df_decimals u_digits')
GroomedUncertainComplex = namedtuple(
    'ucomplex','x u r df label precision df_decimals re_u_digits im_u_digits'
)






"""
Provides LU decomposition functions for Python objects stored in 2D arrays.

:func:`solve` returns the solution vector `x` to the equation `a.x = b`
:func:`ludet` returns the determinant of a previously LU-decomposed matrix 
:func:`ludcmp` performs LU decomposition
:func:`invab` 
"""
from __future__ import division
from functools import reduce
try:
    xrange  # Python 2
except NameError:
    xrange = range

import numpy as np 
import array

__all__ = (
    'solve'
,   'invab'
,   'ludet'
,   'ludcmp'
)

# -------------------------------------------------------------
def ludcmp(a):
    """
    Return (a_lu,idx,parity), where 'a_lu' is `a` decomposed
    
    The array `a` is decomposed by row-wise LU decomposition;
    `idx` records the row permutations from partial pivoting,
    `parity` is the parity of row switching (+1 or -1, for even or odd)

    .. note:: 
        The array `a` is assigned the decomposition values.
    
    .. note:: 
        An implementation of :func:`abs()` must be defined 
        for the type of the array elements.

    Parameters
    ----------
    a : :class:`UncertainArray`
    
    """
    if len(a.shape) != 2:
        raise RuntimeError( 
            "A 2D array is needed, got  shape={}".format(a.shape)
        )
    elif a.shape[0] != a.shape[1]:
        raise RuntimeError( 
            "A square array is needed, shape={}".format(a.shape)
        )
        
    N = a.shape[0]

    indx = array.array('l',[0] * N)   # Holds indices
    vv = array.array('d',[0.0] * N)   
    parity = 1

    for i in xrange(N):
        big = 0.0
        for j in xrange(N):
            temp = abs( a[i,j] )
            big = temp if temp > big else big
                
        try:
            vv[i] = 1.0/big
        except ZeroDivisionError:
            raise RuntimeError('zero column')
            
    for j in xrange(N):
        for i in xrange(j):
            a[i,j] = reduce(
                lambda sum,k: sum - (a[i,k] * a[k,j]), 
                    xrange(i),  # for k in xrange(i)
                    a[i,j]      # initial value
            )

        big = 0.0
        for i in xrange(j,N):
            a[i,j] = reduce(
                lambda sum,k: sum - (a[i,k] * a[k,j]), 
                    xrange(j),  # for k in xrange(j)
                    a[i,j]      # initial value
            )
            dum = vv[i] * abs(a[i,j])
            if(dum >= big):
                big = dum
                imax = i
            
        if(j != imax):
            for k in xrange(N):
                a[imax,k], a[j,k] = a[j,k], a[imax,k]
            parity *= -1
            vv[imax] = vv[j]
            
        indx[j] = imax
        if( a[j,j] == 0.0 ):
            raise RuntimeError("singular pivot element")
        
        if(j != N-1):
            tmp = 1.0 / a[j,j]
            for i in xrange(j+1,N):
                a[i,j] *= tmp

    return (a,indx,parity)

#----------------------------------------------------------------------------
def _lubksb(a_lu,idx,b):
    """
    LU back-substitution 

    The argument `a_lu` is the LU decomposition of `a`,
    where `a` is a 2D array (`x` and `b` are 1D).
    
    The argument `a_lu` is not changed, so the routine 
    may be used successively on different cases of `b`.
    
    The contents of `b` are changed by the function:
    `b` becomes the solution vector.
    
    The argument `b` need only be a sequence type;
    `a_lu` must be array-like.

    Usage::
        a_lu, idx = _ludcmp(a)
        x = _lubksub( a_lu, idx, b )

    """
    
    # If `b` does not contain the same type as `a_lu` 
    # the algorithm can calculate unexpected results! 
    # It is necessary to try to harmonize them in 
    # the calling context.
            
    N = a_lu.shape[0]

    ii = -1
    for i in xrange(N):
        ip = idx[i]
        sum, b[ip] = b[ip], b[i]
        if( ii != -1 ):
            for j in xrange(ii,i):
                sum -= a_lu[i,j] * b[j]
        elif(sum != 0.0):
            ii = i
        b[i] = sum

    for i in xrange(N-1,-1,-1):
        b[i] = reduce(
            lambda sum,j: sum - a_lu[i,j]*b[j],
            xrange(i+1,N),
            b[i]
        )/a_lu[i,i]
            
    return b

#---------------------------------------------------------
def invab(a,b):
    """
    Return the solution `x` to the linear equation`a.x = b` when `b` is 2D
    
    `a` and `b` must be array-like
    Neither `a` or `b` are changed by the function 
    
    """
    # Required by _lubksb
    assert a.dtype == b.dtype,\
        "{} != {}".format(a.dtype, b.dtype)
        
    if len(b.shape) != 2:
        raise RuntimeError( 
            "A 2D array is needed for `b`, got  shape={}".format(b.shape)
        )
    # `a` shape is checked when `ludcmp` is called 
    a_lu,idx,p = ludcmp(a.copy())

    N = a_lu.shape[0]
    M = b.shape[1]
        
    y = np.empty( (N,M), a.dtype )

    # For each column in matrix `b`
    for j in xrange(M):
    
        # TODO: consider a slice here
        col = _lubksb(a_lu,idx,[ b[i,j] for i in xrange(N) ])
        
        # TODO: consider a slice here
        for i in xrange(N):
            y[i,j] = col[i]
            
    return y

# -----------------------------------------------------------------
def ludet(a_lu,p):
    """
    Return the determinant of a matrix `a` using its LU decomposition `a_lu`

    Usage::
        a = .... # some matrix
        parity = 1 # value is not important
        a_lu,i,p = ludcmp(a.copy())
        d = ludet(a_lu,p)

    """
    Nx,Ny = a_lu.shape
    if Nx != Ny:
        raise RuntimeError(
            "matrix must be square, got {}".format(a_lu.shape)
        )
    elif p not in (1,-1):
        raise RuntimeError(
            "parity must be +/- 1, got {!r}".format(p)
        )
        
    return reduce(
        lambda p,i: p * a_lu[i,i],
        xrange(Nx),
        p
    )

#------------------------------------------------
def solve(a,b):
    """
    Return the solution `x` for `a.x = b` 
    
    Both `a` and `b` must be array-like
    `a` is a square 2D array, `b` is a 1D array
    Neither `a` or `b` are changed by the function 

    """
    # Required by _lubksb
    assert a.dtype == b.dtype,\
        "{} != {}".format(a.dtype, b.dtype)
        
    a_lu,i,p = ludcmp( a.copy() )
    return _lubksb( a_lu,i,b.copy() )


    